!+
! Module: specadd_mod
!
! Authors:  M. Ceballos - IFCA
!            F.Carrera -  IFCA
!            X. Barcons -  IFCA
!          
! Description: Add input spectra into a multiple output spectrum
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module specadd_mod

  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

  public specadd

contains 
  
  !---------------------------------------------------------------------
  subroutine specadd
  !---------------------------------------------------------------------
    

    implicit none


    real(kind=single) :: rAtt
    real(kind=single), dimension(:), pointer :: pColBgs,pColCorrs,pColExpos
    real(kind=single), dimension(:), pointer :: pColStatInSpec
    real(kind=single), dimension(:), pointer :: pr32ColDataInSpec
    real(kind=single),dimension(:),allocatable :: Stat,r32Data


    integer,parameter :: dimsChar(0:0)=80
    integer(kind=Int16), dimension(:), pointer :: pColNum, &
         pColChanInSpec, pColGrpInSpec, pColQualInSpec
    integer(kind=Int16),dimension(:), allocatable :: Chan,Qual,Grp,nelemvl
    integer(kind=Int32), dimension(:), allocatable :: i32Data
    integer(kind=Int32), dimension(:), pointer :: pi32ColDataInSpec
    integer(kind=Int32) :: irow,nchans,iAtt,nspecs,i,id,nrowsMulSpec
    integer :: initRow,finalRow,nOfElem,typeDataInSpec
    integer,dimension(0) :: zeroSize
    integer,dimension(:),pointer :: dimCol


    character(len=1000),dimension(:), allocatable :: inSpec 
    character(len=1000) :: outSpec
    character(len=80):: strAtt
    character(len=1024) :: rid
    character(len=6) ::temp
    character(len=10),dimension(4) :: columnName
    
    logical :: useold, hQual, hGrp, false=.false.
    logical :: hStatE=.false.,realData=.false.,realDataMult=.false.

    type(DataSetT) :: setInSpec, setOutSpec
    type(TableT) :: tabInSpec,tabOutSpec
    type(ColumnT) :: colChanInSpec,colDataInSpec,colStatInSpec
    type(ColumnT) :: colQualInSpec, colGrpInSpec
    type(ColumnT) :: colNum, colChan, colData, colStat, colBgs
    type(ColumnT) :: colBgf, colCorrs, colCorrf,colExpos
    type(ColumnT) :: colRowid, colRespf, colArf, colQual, colGrp

!------------------------------------------
!   read input parameters from param file
!------------------------------------------

    nspecs=parameterCount("sets")    

    allocate(inSpec(1:nspecs+1))
    outSpec=stringParameter("sumset")

    write(errstr,*)'Reading ',nspecs,' single spectrum(a)...'
    call message(AppMsg,Verbose)
    do i=1,nspecs
       inSpec(i)=stringParameter("sets",i-1)
       setInSpec=dataset(inSpec(i),Read)

       write(errstr,*)'single spectrum:',trim(inSpec(i)),' read.'
       call message(AppMsg,Verbose)

       if(hasBlock(setInSpec,"SPECTRUM") .EQV. false)then
          write(errstr,*)"Error: Not an spectral file:",inSpec(i)
          call error('NotSpecFile')
       endif
       tabInSpec=table(setInSpec,"SPECTRUM")
       if(hasColumn(tabInSpec,"SPEC_NUM") .NEQV. false) then
          write(errstr,*)"Error: Not a single spectral file:",inSpec(i)
          call error('NotSingleSpecFile')
       endif
       if(hasColumn(tabInSpec,"STAT_ERR"))then
          hStatE=.true.
       endif
       nchans=numberOfRows(tabInSpec)
       colDataInSpec=column(tabInSpec,"COUNTS")
       typeDataInSpec=dataType(colDataInSpec)
       if(typeDataInSpec == REAL32)then
          RealData=.true.
       endif
       call release(tabInSpec)
       call release(setInSpec)
    enddo

    useold=booleanParameter("useoldsumset")

    !--------------------------------------------------------------
    !   Create output: multiple spectrum FITS file
    !--------------------------------------------------------------
    
    if(useold .EQV. false)then
       setOutSpec=dataSet(outSpec,Create)
    endif

    !-------------------------------------------------------------
    ! If there exists a previous multiple spectral file --> copy it
    ! before new files are added to output file.
    !-------------------------------------------------------------


    if(useold)then
       inSpec(nspecs+1)=stringParameter("oldsumset")
       setInSpec=dataset(inSpec(nspecs+1),Read)
       tabInSpec=table(setInSpec,"SPECTRUM")
       nrowsMulSpec=numberOfRows(tabInSpec)

       write(errstr,*)'Read multiple spectrum called:',trim(inSpec(nspecs+1))
       call message(AppMsg,Verbose)

       if(hasColumn(tabInSpec,"SPEC_NUM") .EQV. false)then
          write(errstr,*)"Error: Not a multiple spectral file:",&
               inSpec(nspecs+1)
          call error('NotMultipleSpecFile')
       endif
       call copyDataSet(inSpec(nspecs+1),outSpec)
       setOutSpec=dataset(outSpec,Modify)
       tabOutSpec=Table(setOutSpec,"SPECTRUM")


       !------------------------------------------------------------------
       ! Check existence of mandatory columns in input multiple spectral 
       ! file. If they exist but are not variable-length columns (CHANNEL,
       ! COUNTS,STAT_ERR) delete them and create them again
       !------------------------------------------------------------------

       columnName(1)="SPEC_NUM"
       columnName(2)="CHANNEL"
       columnName(3)="COUNTS"

       do i=1,3
          if(hasColumn(tabOutSpec,trim(columnName(i))) .EQV. false)then
             write(errstr,*)"Column not present in multiple spectral file:",&
               trim(columnName(i))
             call error('ColumnNotPresent')
          endif
       enddo

            ! CHANNEL
       nOfElem=-1
       colChanInSpec=column(tabInSpec,"CHANNEL",Read)
       nOfElem=numberOfElements(colChanInSpec)
       
       call message(AppMsg,Verbose,'Read colum CHANNEL from multiple spectrum')

       if(nOfElem > 0)then !fixed length vector col. --> variable length
          call message(AppMsg,Verbose,'Converting CHANNEL &
               &column to variable length')          
          colChan=column(tabOutSpec,"CHANNEL",Modify)
          call deleteColumn(tabOutSpec,"CHANNEL")
          colChan=addColumn(tabOutSpec,"CHANNEL",Integer16,"",&
               dimensions=zeroSize)
          ! read size of fixed length column
          dimCol=>dimensions(colChanInSpec)
          do irow=0,nrowsMulSpec-1
             allocate(Chan(1:dimCol(0)))
             pColChanInSpec => int16Data(colChanInSpec,irow)
             Chan=pColChanInSpec
             call setData(colChan,irow,Chan(1:dimCol(0)))
             deallocate(Chan)
          enddo
          call release(colChan)
       endif
       call release(colChanInSpec)

       ! Read number of elements of var.len. column CHANNEl just in case
       ! they are needed to add more var.len.columns in the rows already
       ! present in the oldsumset
!       colChan=column(tabOutSpec,"CHANNEL",Modify)
       colChan=column(tabOutSpec,"CHANNEL",Read)
       allocate(nelemvl(0:nrowsMulSpec-1))
       do irow=0,nrowsMulSpec-1
          nelemvl(irow)=numberOfElements(colChan,irow)
       enddo
       call release(colChan)

            ! COUNTS
       nOfElem=-1
       colDataInSpec=column(tabInSpec,"COUNTS",Read)
       nOfElem=numberOfElements(colChanInSpec)
       typeDataInSpec=dataType(colDataInSpec)
       select case(typeDataInSpec)
       case(INTEGER32)
          realDataMult=.false.
       case(REAL32)
          realDataMult=.true.
       case DEFAULT
          call error('InvalidDataType','Invalid Data type&
               & in column DATA in Input Spectrum')
       end select
       
       call message(AppMsg,Verbose,'Read colum COUNTS from multiple spectrum')

       if(nOfElem > 0)then !fixed length vector column --> variable length
          call message(AppMsg,Verbose,'Converting COUNTS &
               &column to variable length')
          colData=column(tabOutSpec,"COUNTS",Modify)
          call deleteColumn(tabOutSpec,"COUNTS")
          if(realDataMult .or. realData)then
             colData=addColumn(tabOutSpec,"COUNTS",Real32,"",&
                  dimensions=zeroSize)
          else
             colData=addColumn(tabOutSpec,"COUNTS",Integer32,"",&
                  dimensions=zeroSize)
          end if
          !read COUNTS column from fixed length
          dimCol=>dimensions(colDataInSpec)
          do irow=0,nrowsMulSpec-1
             if(realDataMult)then
                pr32ColDataInSpec => real32Data(colDataInSpec,irow)
                allocate(r32Data(1:dimCol(0)))
                r32Data=pr32ColDataInSpec                
                call setData(colData,irow,r32Data(1:dimCol(0)))
                deallocate(r32Data)
             elseif(realData)then
                pi32ColDataInSpec => int32Data(colDataInSpec,irow)
                allocate(r32Data(1:dimCol(0)))
                r32Data=real(pi32ColDataInSpec)
                call setData(colData,irow,r32Data(1:dimCol(0)))
                deallocate(r32Data)
             else
                allocate(i32Data(1:dimCol(0)))
                i32Data=pi32ColDataInSpec                
                call setData(colData,irow,i32Data(1:dimCol(0)))
                deallocate(i32Data)
             endif
          enddo
          call release(colData)
       endif
       call release(colDataInSpec)

       ! STAT_ERR
       ! if present in oldsumset, check that it's var.len.vec.col
       ! if present only in single spectrum, add it to multiple set
       ! if not present, write a keyword

       if(hasColumn(tabOutSpec,"STAT_ERR"))then
          nOfElem=-1
          colStatInSpec=column(tabInSpec,"STAT_ERR",Read)
          nOfElem=numberOfElements(colStatInSpec)

          call message(AppMsg,Verbose,'Read colum &
               & STAT_ERR from multiple spectrum')

          if(nOfElem > 0)then !fixed length vector column --> variable length
             call message(AppMsg,Verbose,'Converting STAT_ERR &
               &column to variable length')
             colStat=column(tabOutSpec,"STAT_ERR",Modify)
             call deleteColumn(tabOutSpec,"STAT_ERR")
             colStat=addColumn(tabOutSpec,"STAT_ERR",Real32,"",&
                  dimensions=zeroSize)
             dimCol=>dimensions(colStatInSpec)
             do irow=0,nrowsMulSpec-1
                allocate(Stat(1:dimCol(0)))
                pColStatInSpec => Real32Data(colStatInSpec,irow)
                Stat=pColStatInSpec
                call setData(colStat,irow,Stat(1:dimCol(0)))
                deallocate(Stat)
             enddo
             call release(colStat)
          endif
          call release(colStatInSpec)
       elseif(hStatE)then
          colStat=addColumn(tabOutSpec,"STAT_ERR",Real32,"",&
               dimensions=zeroSize)
          call message(AppMsg,Verbose,'Adding column STAT_ERR &
               &to multiple spectrum')
          do irow=0,nrowsMulSpec-1
             allocate(Stat(1:nelemvl(irow)))
             Stat=0.0
             call setData(colStat,irow,Stat(1:nelemvl(irow)))
             deallocate(Stat)
          enddo
          call release(colStat)
          call release(colChan)
       else
          call setAttribute(attributable(tabOutSpec),"POISSERR",&
               booleanAttribute(tabOutSpec,'POISSERR'),&
               "Poisson errors appropriate")
       endif
       

       !----------------------------------------------------------
       ! Add other columns to output set (those not present in multiple 
       ! spectral file)
       !----------------------------------------------------------
       
       
       
       ! QUALITY - optional (mandatory keyword)
       ! If present, check var len column
       
       if(hascolumn(tabOutSpec,"QUALITY") .EQV. false)then
          colQual=addColumn(tabOutSpec,"QUALITY",Integer16,"",&
               dimensions=zeroSize)
          call message(AppMsg,Verbose,'Adding column QUALITY &
               &to multiple spectrum')
          if(hasAttribute(tabInSpec,"QUALITY"))then
             iAtt=int16Attribute(tabInSpec,"QUALITY")
          else
             iAtt=0
          endif
          do irow=0,nrowsMulSpec-1
             allocate(Qual(1:nelemvl(irow)))
             Qual=iAtt
             call setData(colQual,irow,Qual(1:nelemvl(irow)))
             deallocate(Qual)
          enddo
          call release(colQual)
       else
          nOfElem=-1
          colQualInSpec=column(tabInSpec,"QUALITY",Read)
          nOfElem=numberOfElements(colQualInSpec)

          call message(AppMsg,Verbose,'Read colum QUALITY &
               &from multiple spectrum')

          if(nOfElem > 0)then !fixed length vector column --> variable length
          call message(AppMsg,Verbose,'Converting QUALITY &
               &column to variable length')
             colQual=column(tabOutSpec,"QUALITY",Modify)
             call deleteColumn(tabOutSpec,"QUALITY")
             colQual=addColumn(tabOutSpec,"QUALITY",Integer16,"",&
                  dimensions=zeroSize)
             dimCol=>dimensions(colQualInSpec)
             do irow=0,nrowsMulSpec-1
                allocate(Qual(1:dimCol(0)))
                pColQualInSpec => Int16Data(colQualInSpec,irow)
                Qual=pColQualInSpec
                call setData(colQual,irow,Qual(1:dimCol(0)))
                deallocate(Qual)
             enddo
             call release(colQual)
          endif
          call release(colQualInSpec)
       endif



       ! GROUPING - optional (mandatory keyword)
       
       if(hascolumn(tabOutSpec,"GROUPING") .EQV. false)then
          colGrp=addColumn(tabOutSpec,"GROUPING",Integer16,"",&
               dimensions=zeroSize)
          call message(AppMsg,Verbose,'Adding column GROUPING &
               &to multiple spectrum')
          iAtt=1
          do irow=0,nrowsMulSpec-1
             allocate(Grp(1:nelemvl(irow)))
             Grp=iAtt
             call setData(colGrp,irow,Grp(1:nelemvl(irow)))
             deallocate(Grp)
          enddo
          call release(colGrp)
       else
          nOfElem=-1
          colGrpInSpec=column(tabInSpec,"GROUPING",Read)
          nOfElem=numberOfElements(colGrpInSpec)

          call message(AppMsg,Verbose,'Read colum GROUPING &
               &from multiple spectrum')

          if(nOfElem > 0)then !fixed length vector column --> variable length
          call message(AppMsg,Verbose,'Converting GROUPING &
               &column to variable length')
             colGrp=column(tabOutSpec,"GROUPING",Modify)
             call deleteColumn(tabOutSpec,"GROUPING")
             colGrp=addColumn(tabOutSpec,"GROUPING",Integer16,"",&
                  dimensions=zeroSize)
             dimCol=>dimensions(colGrpInSpec)
             do irow=0,nrowsMulSpec-1
                allocate(Grp(1:dimCol(0)))
                pColGrpInSpec => Int16Data(colGrpInSpec,irow)
                Grp=pColGrpInSpec
                call setData(colGrp,irow,Grp(1:dimCol(0)))
                deallocate(Grp)
             enddo
             call release(colGrp)
          endif
          call release(colGrpInSpec)

       endif


       ! ROWID - optional
       if(hascolumn(tabOutSpec,"ROWID") .EQV. false)then
          colRowid=addColumn(tabOutSpec,"ROWID",String," ",dimsChar)
          call message(AppMsg,Verbose,'Adding column ROWID &
            &to multiple spectrum')

          do irow=0,nrowsMulSpec-1
             id=irow+1
             write(temp(1:5),"('(',(I3),')')")id
             rid=trim(inSpec(nspecs+1))//trim(temp)
             call setStringCell(colRowid,irow,trim(rid))
          enddo
          call release(colRowid)
       endif
       


       ! BACKFILE - optional (mandatory keyword)

       if(hascolumn(tabOutSpec,"BACKFILE") .EQV. false)then
          colBgf=addColumn(tabOutSpec,"BACKFILE",String,"",dimsChar)
          call message(AppMsg,Verbose,'Adding column BACKFILE &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"BACKFILE"))then
             strAtt=stringAttribute(tabInSpec,"BACKFILE")
             if(trim(strAtt)=="")strAtt="NONE"
          else
             strAtt="NONE"
          endif
          do irow=0,nrowsMulSpec-1
             call setStringCell(colBgf,irow,strAtt)
          enddo
          call release(colBgf)
       endif



       ! BACKSCAL - optional (mandatory keyword)

       if(hascolumn(tabOutSpec,"BACKSCAL") .EQV. false)then
          colBgs=addColumn(tabOutSpec,"BACKSCAL",Real32,"")
          call message(AppMsg,Verbose,'Adding column BACKSCAL &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"BACKSCAL"))then
             rAtt=real32Attribute(tabInSpec,"BACKSCAL")
          else
             rAtt=1.0
          endif
          pColBgs => real32Data(colBgs)
          do irow=0,nrowsMulSpec-1
             pColBgs(irow)=rAtt
          enddo
          call release(colBgs)
       endif



       ! CORRFILE - optional (mandatory keyword)

       if(hascolumn(tabOutSpec,"CORRFILE") .EQV. false)then
          colCorrf=addColumn(tabOutSpec,"CORRFILE",String,"",dimsChar)
          call message(AppMsg,Verbose,'Adding column CORRFILE &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"CORRFILE"))then
             strAtt=stringAttribute(tabInSpec,"CORRFILE")
             if(trim(strAtt)=="")strAtt="NONE"
          else
             strAtt="NONE"
          endif
          do irow=0,nrowsMulSpec-1
             call setStringCell(colCorrf,irow,strAtt)
          enddo
          call release(colCorrf)
       endif



       ! CORRSCAL - optional (mandatory keyword)

       if(hascolumn(tabOutSpec,"CORRSCAL") .EQV. false)then
          colCorrs=addColumn(tabOutSpec,"CORRSCAL",Real32,"")
          call message(AppMsg,Verbose,'Adding column CORRSCAL &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"CORRSCAL"))then
             rAtt=real32Attribute(tabInSpec,"CORRSCAL")
          else
             rAtt=1.0
          endif
          pColCorrs => real32Data(colCorrs)
          do irow=0,nrowsMulSpec-1
             pColCorrs(irow)=rAtt          
          enddo
          call release(colCorrs)
       endif

       ! RESPFILE - optional (mandatory keyword)
       
       if(hascolumn(tabOutSpec,"RESPFILE") .EQV. false)then
          colRespf=addColumn(tabOutSpec,"RESPFILE",String,"",dimsChar)
          call message(AppMsg,Verbose,'Adding column RESPFILE &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"RESPFILE"))then
             strAtt=stringAttribute(tabInSpec,"RESPFILE")
             if(trim(strAtt)=="")strAtt="NONE"
          else
             strAtt="NONE"
          endif
          do irow=0,nrowsMulSpec-1
             call setStringCell(colRespf,irow,strAtt)
          enddo
          call release(colRespf)
       endif

       ! ANCRFILE - optional (mandatory keyword)
       
       if(hascolumn(tabOutSpec,"ANCRFILE") .EQV. false)then
          colArf=addColumn(tabOutSpec,"ANCRFILE",String,"",dimsChar)
          call message(AppMsg,Verbose,'Adding column ANCRFILE &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"ANCRFILE"))then
             strAtt=stringAttribute(tabInSpec,"ANCRFILE")
             if(trim(strAtt)=="")strAtt="NONE"
          else
             strAtt="NONE"
          endif
          do irow=0,nrowsMulSpec-1
             call setStringCell(colArf,irow,strAtt)
          enddo
          call release(colArf)
       endif

       ! EXPOSURE - optional (XSPEC mandatory keyword/column)

       if(hascolumn(tabOutSpec,"EXPOSURE") .EQV. false)then
          colExpos=addColumn(tabOutSpec,"EXPOSURE",Real32,"")
          call message(AppMsg,Verbose,'Adding column EXPOSURE &
            &to multiple spectrum')
          if(hasAttribute(tabInSpec,"EXPOSURE"))then
             rAtt=real32Attribute(tabInSpec,"EXPOSURE")
          else
             write(errstr,*)'Exposure Attribute not found in spec:',&
                  trim(inSpec(i)),' Taking exposure value : 1.00E+00 s'
             call warning('NoExposure')
             rAtt=1.0
          endif
          pColExpos => real32Data(colExpos)
          do irow=0,nrowsMulSpec-1
             pColExpos(irow)=rAtt          
          enddo
          call release(colExpos)
       endif



       !----------------------------------------------
       ! Remove attributes that will be now in columns
       ! (just in case they have been copied)
       !----------------------------------------------
       if(hasAttribute(tabOutSpec,"BACKFILE"))&
            call deleteAttribute(tabOutSpec,"BACKFILE")
       if(hasAttribute(tabOutSpec,"BACKSCAL"))&
            call deleteAttribute(tabOutSpec,"BACKSCAL")
       if(hasAttribute(tabOutSpec,"CORRFILE"))&
            call deleteAttribute(tabOutSpec,"CORRFILE")
       if(hasAttribute(tabOutSpec,"CORRSCAL"))&
            call deleteAttribute(tabOutSpec,"CORRSCAL")
       if(hasAttribute(tabOutSpec,"ANCRFILE"))&
            call deleteAttribute(tabOutSpec,"ANCRFILE")
       if(hasAttribute(tabOutSpec,"RESPFILE"))&
            call deleteAttribute(tabOutSpec,"RESPFILE")
       if(hasAttribute(tabOutSpec,"QUALITY"))&
            call deleteAttribute(tabOutSpec,"QUALITY")
       if(hasAttribute(tabOutSpec,"GROUPING"))&
            call deleteAttribute(tabOutSpec,"GROUPING")
       if(hasAttribute(tabOutSpec,"EXPOSURE"))&
            call deleteAttribute(tabOutSpec,"EXPOSURE")
       
       call release(tabInSpec)
       call release(setInSpec)

       !--------------------------------------------
       ! Insert n(=number of single spectra) rows
       !--------------------------------------------
       
       call message(AppMsg,Verbose,'Inserting rows to store single spectra')

       call insertRows(tabOutSpec,nrowsMulSpec,nspecs)

       !----------------------------------------------------
       ! Modify output set to host new input single
       ! spectral files 
       !----------------------------------------------------

       colNum=column(tabOutSpec,"SPEC_NUM",Modify)
       colChan=column(tabOutSpec,"CHANNEL",Modify)
       colData=column(tabOutSpec,"COUNTS",Modify)
       if (hasColumn(tabOutSpec,"STAT_ERR"))&
            colStat=column(tabOutSpec,"STAT_ERR",Modify)
       colQual=column(tabOutSpec,"QUALITY",Modify)
       colGrp=column(tabOutSpec,"GROUPING",Modify)
       colRowid=column(tabOutSpec,"ROWID",Modify)
       colBgf=column(tabOutSpec,"BACKFILE",Modify)
       colBgs=column(tabOutSpec,"BACKSCAL",Modify)
       colCorrf=column(tabOutSpec,"CORRFILE",Modify)
       colCorrs=column(tabOutSpec,"CORRSCAL",Modify)
       colRespf=column(tabOutSpec,"RESPFILE",Modify)
       colArf=column(tabOutSpec,"ANCRFILE",Modify)
       colExpos=column(tabOutSpec,"EXPOSURE",Modify)


       !------------------------------------------------
       ! Pointers to non-vector (and non-string) columns
       !------------------------------------------------
       pColNum => int16Data(colNum)
       pColCorrs => real32Data(colCorrs)
       pColBgs => real32Data(colBgs)
       pColExpos => real32Data(colExpos)

       initRow=nrowsMulSpec
       finalRow=nrowsMulSpec+nspecs-1

    else

       call message(AppMsg,Verbose,'Creating new table SPECTRUM &
            & in multiple spectrum')

       tabOutSpec=addTable(setOutSpec,"SPECTRUM",nspecs)
       
       !------------------------------------------------------------
       ! Add columns to output set 
       ! 
       !------------------------------------------------------------
       
       ! SPEC_NUM - mandatory
       
       colNum=addColumn(tabOutSpec,"SPEC_NUM",Integer16," ")
       call message(AppMsg,Verbose,'Adding column SPEC_NUM &
            &to new multiple spectrum')

       ! CHANNEL - mandatory
       colChan=addColumn(tabOutSpec,"CHANNEL",Integer16,"",&
            dimensions=zeroSize)
       call message(AppMsg,Verbose,'Adding column CHANNEL &
            &to new multiple spectrum')
       
       ! DATA - mandatory
       if(realData)then
          colData=addColumn(tabOutSpec,"COUNTS",Real32,"counts",&
            dimensions=zeroSize)
       else
          colData=addColumn(tabOutSpec,"COUNTS",Integer32,"counts",&
            dimensions=zeroSize)
       endif
       call message(AppMsg,Verbose,'Adding column DATA &
            &to new multiple spectrum')

       
       ! STAT - optional
       ! if present in single sets, add to newly created mult. set
       ! otherwise, keep it as POISSERR keyword
       if(hStatE)then
          colStat=addColumn(tabOutSpec,"STAT_ERR",Real32,"counts",&
            dimensions=zeroSize)
          call message(AppMsg,Verbose,'Adding column STAT_ERR &
               &to new multiple spectrum')
       else
          call setAttribute(attributable(tabOutSpec),"POISSERR","T",&
               "Poisson errors appropriate")
       endif
          


       ! QUALITY - optional (mandatory keyword)
       colQual=addColumn(tabOutSpec,"QUALITY",Integer16,"",&
            dimensions=zeroSize)
       call message(AppMsg,Verbose,'Adding column QUALITY &
            &to new multiple spectrum')

       
       ! GROUPING - optional (mandatory keyword)
       colGrp=addColumn(tabOutSpec,"GROUPING",Integer16,"",&
            dimensions=zeroSize)
       call message(AppMsg,Verbose,'Adding column GROUPING &
            &to new multiple spectrum')

       
       ! ROWID - optional
       colRowid=addColumn(tabOutSpec,"ROWID",String," ",dimsChar)
       call message(AppMsg,Verbose,'Adding column ROWID &
            &to new multiple spectrum')

       
       ! BACKFILE - optional (mandatory keyword)
       colBgf=addColumn(tabOutSpec,"BACKFILE",String,"",dimsChar)
       call message(AppMsg,Verbose,'Adding column BACKFILE &
            &to new multiple spectrum')

       
       ! BACKSCAL - optional (mandatory keyword)
       colBgs=addColumn(tabOutSpec,"BACKSCAL",Real32,"")
       call message(AppMsg,Verbose,'Adding column BACKSCAL &
            &to new multiple spectrum')

       
       ! CORRFILE - optional (mandatory keyword)
       colCorrf=addColumn(tabOutSpec,"CORRFILE",String,"",dimsChar)
       call message(AppMsg,Verbose,'Adding column CORRFILE &
            &to new multiple spectrum')

       
       ! CORRSCAL - optional (mandatory keyword)
       colCorrs=addColumn(tabOutSpec,"CORRSCAL",Real32,"")
       call message(AppMsg,Verbose,'Adding column CORRSCAL &
            &to new multiple spectrum')

       
       ! RESPFILE - optional (mandatory keyword)
       colRespf=addColumn(tabOutSpec,"RESPFILE",String,"",dimsChar)
       call message(AppMsg,Verbose,'Adding column RESPFILE &
            &to new multiple spectrum')

       
       ! ANCRFILE - optional (mandatory keyword)
       colArf=addColumn(tabOutSpec,"ANCRFILE",String,"",dimsChar)
       call message(AppMsg,Verbose,'Adding column ANCRFILE &
            &to new multiple spectrum')


       ! EXPOSURE - optional (XSPEC mandatory keyword)
       colExpos=addColumn(tabOutSpec,"EXPOSURE",Real32,"")
       call message(AppMsg,Verbose,'Adding column EXPOSURE &
            &to new multiple spectrum')
       
       !
       ! Pointer access to non vector columns
       !
       
       pColNum => int16Data(colNum)
       pColCorrs => real32Data(colCorrs)
       pColBgs => real32Data(colBgs)
       pColExpos => real32Data(colExpos)
       
       initRow=0
       finalRow=nspecs-1
    endif
       
    
    i=1 !initial input spectrum
    do irow=initRow,finalRow
       write(errstr,*)'Row number ',irow,' in multiple spectrum'
       call message(AppMsg,Verbose)
       write(errstr,*)'Storing spectrum ',trim(inSpec(i))
       call message(AppMsg,Verbose)

       setInSpec=dataset(inSpec(i),Read)
       tabInSpec=table(setInSpec,"SPECTRUM")
       nchans=numberOfRows(tabInSpec)
       allocate(Chan(nchans))
       allocate(Qual(nchans))
       allocate(Grp(nchans))

       if(useold.EQV. false)then
          call setAttributes(attributable(setOutSpec),&
               attributable(setInSpec))
          call setAttributes(attributable(tabOutSpec),&
               attributable(tabInSpec))
          ! remove attributes that will be in columns
          
          if(hasAttribute(tabOutSpec,"BACKFILE"))&
               call deleteAttribute(tabOutSpec,"BACKFILE")
          if(hasAttribute(tabOutSpec,"BACKSCAL"))&
               call deleteAttribute(tabOutSpec,"BACKSCAL")
          if(hasAttribute(tabOutSpec,"CORRFILE"))&
               call deleteAttribute(tabOutSpec,"CORRFILE")
          if(hasAttribute(tabOutSpec,"CORRSCAL"))&
               call deleteAttribute(tabOutSpec,"CORRSCAL")
          if(hasAttribute(tabOutSpec,"ANCRFILE"))&
               call deleteAttribute(tabOutSpec,"ANCRFILE")
          if(hasAttribute(tabOutSpec,"RESPFILE"))&
               call deleteAttribute(tabOutSpec,"RESPFILE")
          if(hasAttribute(tabOutSpec,"QUALITY"))&
               call deleteAttribute(tabOutSpec,"QUALITY")
          if(hasAttribute(tabOutSpec,"GROUPING"))&
               call deleteAttribute(tabOutSpec,"GROUPING")
          if(hasAttribute(tabOutSpec,"EXPOSURE"))&
               call deleteAttribute(tabOutSpec,"EXPOSURE")
       endif
       !--------------------------------------------------------
       !   fill in a new row (irow) in the output table columns:
       !--------------------------------------------------------       

       !   Column: SPEC_NUM
       !--------------------
       pColNum(irow)=irow+1

       !   Columns: CHANNEL, COUNTS
       !----------------------------
       colChanInSpec=column(tabInSpec,"CHANNEL",read)
       pColChanInSpec => int16Data(colChanInSpec)
       Chan=pColChanInSpec
       call setData(colChan,irow,Chan(1:nchans))
       call release(colChanInSpec)
       colDataInSpec=column(tabInSpec,"COUNTS",read)
       typeDataInSpec=dataType(colDataInSpec)

       if(typeDataInSpec == INTEGER32)then
          pi32ColDataInSpec => int32Data(colDataInSpec)
          if(realDataMult .or. realData)then
             allocate(r32Data(nchans))
             r32Data=real(pi32ColDataInSpec)
             call setData(colData,irow,r32Data(1:nchans))
          else
             allocate(i32Data(nchans))
             i32Data=pi32ColDataInSpec
             call setData(colData,irow,i32Data(1:nchans))
          endif
       else
          pr32ColDataInSpec => real32Data(colDataInSpec)
          allocate(r32Data(nchans))
          r32Data=real(pi32ColDataInSpec)
          call setData(colData,irow,r32Data(1:nchans))
       endif
       call release(colDataInSpec)

       !   Column: STAT_ERR
       !--------------------
       
       if(hasColumn(tabInSpec,"STAT_ERR"))then 
          colStatInSpec=column(tabInSpec,"STAT_ERR",read)
          pColStatInSpec => real32Data(colStatInSpec)
          allocate(Stat(nchans))
          Stat=pColStatInSpec
          call release(colStatInSpec)
          call setData(colStat,irow,Stat(1:nchans))
          if(hasAttribute(tabOutSpec,"POISSERR"))&
               call deleteAttribute(tabOutSpec,"POISSERR")
          deallocate(Stat)
       else if(hasColumn(tabOutSpec,"STAT_ERR"))then
          allocate(Stat(nchans))
          Stat=0.0
          call setData(colStat,irow,Stat(1:nchans))
          if(hasAttribute(tabOutSpec,"POISSERR"))&
               call deleteAttribute(tabOutSpec,"POISSERR")
          deallocate(Stat)
       endif

       
       !   Column: QUALITY
       !-------------------
       
       hQual=hasColumn(tabInSpec,"QUALITY")
       if(hQual)then
          colQualInSpec=column(tabInSpec,"QUALITY",read)
          pColQualInSpec => int16Data(colQualInSpec)
          Qual=pColQualInSpec
          call release(colQualInSpec)
       else
          if(hasAttribute(tabInSpec,"QUALITY"))then
             iAtt=int16Attribute(tabInSpec,"QUALITY")
          else
             iAtt=0
          endif
          Qual=iAtt
       endif
       call setData(colQual,irow,Qual(1:nchans))
       
       !   Column: GROUPING
       !--------------------
       
       hGrp=hasColumn(tabInSpec,"GROUPING")
       if(hGrp)then
          colGrpInSpec=column(tabInSpec,"GROUPING",read)
          pColGrpInSpec => int16Data(colGrpInSpec)
          Grp=pColGrpInSpec
          call release(colGrpInSpec)
       else
          iAtt=1
          Grp=iAtt
       endif
       call setData(colGrp,irow,Grp(1:nchans))          
       
       !   Column: ROWID
       !-----------------
       call setStringCell(colRowid,irow,inSpec(i))          
       
       !   Column: BACKFILE
       !--------------------
       if(hasAttribute(tabInSpec,"BACKFILE"))then
          strAtt=stringAttribute(tabInSpec,"BACKFILE")
          if(trim(strAtt)=="")strAtt="NONE"
       else
          strAtt="NONE"
       endif
       call setStringCell(colBgf,irow,strAtt)
       
       !   Column: BACKSCAL
       !--------------------
       if(hasAttribute(tabInSpec,"BACKSCAL"))then
          rAtt=real32Attribute(tabInSpec,"BACKSCAL")
          pColBgs(irow)=rAtt
       else
          pColBgs(irow)=1.0
       endif
       
       !   Column: CORRFILE       
       !--------------------
       if(hasAttribute(tabInSpec,"CORRFILE"))then
          strAtt=stringAttribute(tabInSpec,"CORRFILE")
          if(trim(strAtt)=="")strAtt="NONE"
       else
          strAtt="NONE"
       endif
       call setStringCell(colCorrf,irow,strAtt)          
       
       !   Column: CORRSCAL 
       !-------------------
       if(hasAttribute(tabInSpec,"CORRSCAL"))then
          rAtt=real32Attribute(tabInSpec,"CORRSCAL")
          pColCorrs(irow)=rAtt
       else
          pColCorrs(irow)=1.0
       endif
       
       !   Column: RESPFILE
       !--------------------
       if(hasAttribute(tabInSpec,"RESPFILE"))then
          strAtt=stringAttribute(tabInSpec,"RESPFILE")
          if(trim(strAtt)=="")strAtt="NONE"
       else
          strAtt="NONE"
       endif
       call setStringCell(colRespf,irow,strAtt)          
       
       !   Column: ANCRFILE 
       !--------------------
       if(hasAttribute(tabInSpec,"ANCRFILE"))then
          strAtt=stringAttribute(tabInSpec,"ANCRFILE")
          if(trim(strAtt)=="")strAtt="NONE"
       else
          strAtt="NONE"
       endif
       call setStringCell(colArf,irow,strAtt)
       
       !   Column: EXPOSURE
       !-------------------
       if(hasAttribute(tabInSpec,"EXPOSURE"))then
          rAtt=real32Attribute(tabInSpec,"EXPOSURE")
          pColExpos(irow)=rAtt
       else
          write(errstr,*)'Exposure Attribute not found in spec:',&
               trim(inSpec(i)),' Taking exposure value : 1.00E+00 s'
          call warning('NoExposure')
          pColExpos(irow)=1.0
       endif
       
       deallocate(Chan)
       if(realData .or. realDataMult)then
          deallocate(r32Data)
       else
          deallocate(i32Data)
       endif
       deallocate(Qual)
       deallocate(Grp)
       
       call release(tabInSpec)
       call release(setInSpec)
       
       i=i+1 !next input spectrum

    enddo !irows
    
    ! release handles
    
    call release(tabOutSpec)
    call release(setOutSpec)
  end subroutine specadd
  
end module specadd_mod
