program testcol

  use errorHandling
  use dal

  
  implicit none 

  integer :: nofrows,nofcols

  integer(kind=Int32), dimension(:), pointer :: pcolData1,pcolData2
  integer(kind=Int32), dimension(:), pointer :: pcolData3
  integer, dimension(:),allocatable :: Data
  integer,dimension(4) :: nelems

  logical :: hascol


  type(DataSetT) :: specDataSet,spec1Set,spec2Set,spec3Set
  type(TableT) :: specTable,spec1Table,spec2Table,spec3Table
  type(ColumnT) :: colData,colChan,col1Data,col2Data,col3Data

! Access output Spectral File

  specDataSet=dataSet("sumset.fits",Read)
  specTable=table(specDataSet,"SPECTRUM")


! Output File contains correct dimensions:
!     a)  Number of rows = 4 (one for each single spectrum + 
!                              2 for the combined one)

  nofrows=numberOfRows(specTable)
  if(nofrows < 4) &
       call fatal('IncompOutput','Output file is Incomplete')

!     b)  Number of columns = 14 and  names  are correct

  nofcols=numberOfColumns(specTable)
  if(nofcols < 14)call fatal('IncompOutput','Less than 14 columns')

  hascol=hasColumn(specTable,"SPEC_NUM")
  if(.NOT.hascol)call fatal('IncompOutput','Column SPEC_NUM is not present')

  hascol=hasColumn(specTable,"CHANNEL")
  if(.NOT.hascol)call fatal('IncompOutput','Column CHANNEL is not present')
  colChan=column(specTable,"CHANNEL",read)

  hascol=hasColumn(specTable,"COUNTS")
  if(.NOT.hascol)call fatal('IncompOutput','Column COUNTS is not present')
  colData=column(specTable,"COUNTS",read)

  hascol=hasColumn(specTable,"STAT_ERR")
  if(.NOT.hascol)call fatal('IncompOutput','Column STAT_ERR is not present')

  hascol=hasColumn(specTable,"QUALITY")
  if(.NOT.hascol)call fatal('IncompOutput','Column QUALITY is not present')

  hascol=hasColumn(specTable,"GROUPING")
  if(.NOT.hascol)call fatal('IncompOutput','Column GROUPING is not present')

  hascol=hasColumn(specTable,"ROWID")
  if(.NOT.hascol)call fatal('IncompOutput','Column ROWID is not present')

  hascol=hasColumn(specTable,"BACKSCAL")
  if(.NOT.hascol)call fatal('IncompOutput','Column BACKSCAL is not present')

  hascol=hasColumn(specTable,"BACKFILE")
  if(.NOT.hascol)call fatal('IncompOutput','Column BACKFILE is not present')

  hascol=hasColumn(specTable,"CORRSCAL")
  if(.NOT.hascol)call fatal('IncompOutput','Column CORRSCAL is not present')

  hascol=hasColumn(specTable,"CORRFILE")
  if(.NOT.hascol)call fatal('IncompOutput','Column CORRFILE is not present')

  hascol=hasColumn(specTable,"RESPFILE")
  if(.NOT.hascol)call fatal('IncompOutput','Column RESPFILE is not present')

  hascol=hasColumn(specTable,"ANCRFILE")
  if(.NOT.hascol)call fatal('IncompOutput','Column ANCRFILE is not present')

  hascol=hasColumn(specTable,"EXPOSURE")
  if(.NOT.hascol)call fatal('IncompOutput','Column EXPOSURE is not present')
  
!
! Test some values
!

  spec1Set=dataSet("specm1.fits",Read)
  spec1Table=table(spec1Set,"SPECTRUM")
  spec2Set=dataSet("specpn.fits",Read)
  spec2Table=table(spec2Set,"SPECTRUM")
  spec3Set=dataSet("spec2m1.fits",Read)
  spec3Table=table(spec3Set,"SPECTRUM")


  colData=column(specTable,"COUNTS",Read)
  col1Data=column(spec1Table,"COUNTS",Read)
  col2Data=column(spec2Table,"COUNTS",Read)
  col3Data=column(spec3Table,"COUNTS",Read)

 
  nelems(1)=numberOfRows(col1Data)
  nelems(2)=numberOfRows(col2Data)
  nelems(3)=numberOfRows(col3Data)

  pcolData1=>int32Data(col1Data)
  pcolData2=>int32Data(col2Data)
  pcolData3=>int32Data(col3Data)

  allocate(Data(0:nelems(1)-1))
  Data=int32Data(colData,0)
  if(Data(15) /= pcolData1(15))call fatal('InexactValue','Output file does not reproduce input values in specm1.fits')
  deallocate(Data)

  allocate(Data(0:nelems(2)-1))
  Data=int32Data(colData,1)
  if(Data(7) /= pcolData2(7))call fatal('InexactValue','Output file does not reproduce input values in specpn.fits')
  deallocate(Data)
  
  allocate(Data(0:nelems(3)-1))
  Data=int32Data(colData,2)
  if(Data(100) /= pcolData3(100))call fatal('InexactValue','Output file does not reproduce input values in specpn.fits')
  deallocate(Data)

  call release(colData)
  call release(col1Data)
  call release(col2Data)
  call release(col3Data)

  call release(specTable)
  call release(spec1Table)
  call release(spec2Table)
  call release(spec3Table)

  call release(specDataSet)
  call release(spec1Set)
  call release(spec2Set)
  call release(spec3Set)

end program testcol

